defmodule Liquid.ForElse do
 @moduledoc """
   Like in Shopify's liquid: "For" iterates over an array or collection.
   Several useful variables are available to you within the loop.

   == Basic usage:
      {% for item in collection %}
        {{ forloop.index }}: {{ item.name }}
      {% endfor %}

   == Advanced usage:
      {% for item in collection %}
        <div {% if forloop.first %}class="first"{% endif %}>
          Item {{ forloop.index }}: {{ item.name }}
        </div>
      {% else %}
        There is nothing in the collection.
      {% endfor %}

   You can also define a limit and offset much like SQL.  Remember
   that offset starts at 0 for the first item.

      {% for item in collection limit:5 offset:10 %}
        {{ item.name }}
      {% end %}

    To reverse the for loop simply use {% for item in collection reversed %}

   == Available variables:

   forloop.name:: 'item-collection'
   forloop.length:: Length of the loop
   forloop.index:: The current item's position in the collection;
                   forloop.index starts at 1.
                   This is helpful for non-programmers who start believe
                   the first item in an array is 1, not 0.
   forloop.index0:: The current item's position in the collection
                    where the first item is 0
   forloop.rindex:: Number of items remaining in the loop
                    (length - index) where 1 is the last item.
   forloop.rindex0:: Number of items remaining in the loop
                     where 0 is the last item.
   forloop.first:: Returns true if the item is the first item.
   forloop.last:: Returns true if the item is the last item.
   forloop.parentloop:: Provides access to the parent loop, if present.

"""
  alias Liquid.Render
  alias Liquid.Block
  alias Liquid.Variable
  alias Liquid.Context
  alias Liquid.Expression
  alias Liquid.RangeLookup
  defmodule Iterator do
    defstruct name: nil, collection: nil, item: nil, reversed: false,
                        limit: nil, offset: nil, forloop: %{}
  end

  def syntax, do: ~r/(\w+)\s+in\s+(#{Liquid.quoted_fragment}+)\s*(reversed)?/

  def parse(%Block{nodelist: nodelist}=block, %Liquid.Template{}=t) do
    block = %{block | iterator: parse_iterator(block) }
    case Block.split(block) do
      { true_block, [_,false_block] } ->
        is_blank = Blank.blank?([true_block|false_block])
        { %{block | nodelist: true_block, elselist: false_block, blank: is_blank}, t }
      { _, [] } ->
          is_blank = Blank.blank?(nodelist)
        { %{block | blank: is_blank}, t }
    end
  end

  defp parse_iterator(%Block{markup: markup}) do
    [[_,item|[orig_collection|reversed]]] = Regex.scan(syntax(), markup)
    collection = Expression.parse(orig_collection)
    reversed   = !(reversed |> List.first |> is_nil)
    attributes = Liquid.tag_attributes |> Regex.scan(markup)
    limit      = attributes |> parse_attribute("limit") |> Variable.create
    offset     = attributes |> parse_attribute("offset", "0") |> Variable.create

    %Iterator{name: orig_collection, item: item, collection: collection,
             limit: limit, offset: offset, reversed: reversed}
  end

  defp parse_attribute(attributes, name, default \\ "nil") do
    attributes |> Enum.reduce(default, fn(x, ret) ->
      case x do
        [_, ^name, attribute] when is_binary(attribute) -> attribute
        _ -> ret
      end
    end)
  end

  def render(output, %Block{iterator: it}=block, %Context{}=context) do
    list = parse_collection(it.collection, context)
    list = if is_binary(list) and list != "", do: [list], else: list
    if is_list(list) and Enum.count(list) > 0 do
      list = if it.reversed, do: Enum.reverse(list), else: list
      limit  = lookup_limit(it, context)
      offset = lookup_offset(it, context)
      each(output, [make_ref(), limit,offset], list, block, context)
    else
      Render.render(output, block.elselist, context)
    end
  end

  defp parse_collection(list, _context) when is_list(list), do: list
  defp parse_collection(%Variable{} = variable, context) do
    Variable.lookup(variable, context)
  end

  defp parse_collection(%RangeLookup{} = range, context) do
    RangeLookup.parse(range, context)
  end

  def each(output, _, [], %Block{}=block, %Context{}=context), do: { output, remember_limit(block, context) }
  def each(output, [prev, limit, offset], [h|t]=list, %Block{}=block, %Context{}=context ) do
    forloop = next_forloop(block.iterator, list)
    block   = %{ block | iterator: %{block.iterator | forloop: forloop }}
    assigns = context.assigns |> Map.put("forloop", forloop)
                              |> Map.put(block.iterator.item, h)
    registers = context.registers |> Map.put("changed", {prev,h})

    { output, block_context } = render_content(output, block, %{context | assigns: assigns, registers: registers}, [limit, offset])

    t = if block_context.break == true, do: [], else: t
    each(output, [h, limit, offset], t, block, %{context | assigns: block_context.assigns, registers: block_context.registers})
  end

  defp render_content(output, %Block{iterator: it}=block, context, [limit, offset]) do
    case {should_render?(limit, offset, it.forloop["index"]), block.blank} do
      {true, true} ->
        { _, new_context } = Render.render([], block.nodelist, context)
        { output, new_context }
      {true, _ } ->
        Render.render(output, block.nodelist, context)
      _ ->
        { output, context }
    end
  end

  defp remember_limit(%Block{iterator: it}, context) do
    limit = lookup_limit(it, context) || 0
    remembered = context.offsets[it.name] || 0
    %{ context | offsets: context.offsets |> Map.put(it.name, remembered + limit) }
  end

  defp should_render?(_limit, offset, index) when index <= offset, do: false
  defp should_render?(nil, _, _), do: true
  defp should_render?(limit, offset, index) when index > limit + offset, do: false
  defp should_render?(_limit, _offset, _index), do: true

  defp lookup_limit(%Iterator{limit: limit}, %Context{}=context),
   do: Variable.lookup(limit, context)

  defp lookup_offset(%Iterator{offset: %Variable{name: "continue"}}=it, %Context{}=context),
   do: context.offsets[it.name] || 0

  defp lookup_offset(%Iterator{offset: offset}, %Context{}=context),
   do: Variable.lookup(offset, context)

  defp next_forloop(%Iterator{forloop: loop}=it, count) when map_size(loop) < 1 do
    count = count |> Enum.count
    %{"name" => it.item <> "-" <> it.name,
    "index" => 1,
     "index0" => 0,
     "rindex" => count,
     "rindex0"=> count - 1,
     "length" => count,
     "first"  => true,
     "last"   => count == 1}
  end

  defp next_forloop(%Iterator{forloop: loop}=it, _count) do
    %{"name" => it.item <> "-" <> it.name,
    "index" => loop["index"]  + 1,
     "index0" => loop["index0"] + 1,
     "rindex" => loop["rindex"]  - 1,
     "rindex0"=> loop["rindex0"] - 1,
     "length" => loop["length"],
     "first"  => false,
     "last"   => loop["rindex0"] == 1}
  end

end

defmodule Liquid.Break do
  alias Liquid.Tag, as: Tag
  alias Liquid.Context, as: Context
  alias Liquid.Template, as: Template

  def parse(%Tag{}=tag, %Template{}=template), do: { tag, template }

  def render(output, %Tag{}, %Context{}=context) do
    { output, %{context | break: true } }
  end
end

defmodule Liquid.Continue do
  alias Liquid.Tag, as: Tag
  alias Liquid.Context, as: Context

  def parse(%Tag{}=tag, template), do: { tag, template }

  def render(output, %Tag{}, %Context{}=context) do
    { output, %{context | continue: true } }
  end
end

defmodule Liquid.IfChanged do
  alias Liquid.{Template, Block}

  def parse(%Block{}=block, %Template{}=t), do: { block, t }
  def render(output, %Block{nodelist: nodelist}, context) do
    case context.registers["changed"] do
      {l,r} when l != r -> Liquid.Render.render( output, nodelist, context)
      _ -> {output, context}
    end
  end

end
